from __future__ import annotations
from .Template import MAPPING_TEMPLATE, HOME_DIR
import jinja2
import os
from typing import Type

from lxml import etree

from . import constants, classes


def string_to_bool(text: str) -> bool | None:
    if text == str(True):
        return True
    elif text == str(False):
        return False
    else:
        return None


def build_xml(project: classes.Project) -> etree.ElementTree:
    def add_parent(xml_item: etree.Element, item: classes.Object | classes.PropertySet | classes.Attribute) -> None:
        if item.parent is not None:
            xml_item.set(constants.PARENT, str(item.parent.identifier))
        else:
            xml_item.set(constants.PARENT, constants.NONE)

    def add_predefined_property_sets() -> None:
        xml_grouping = etree.SubElement(xml_project, constants.PREDEFINED_PSETS)
        predefined_psets = [pset for pset in classes.PropertySet if pset.object == None]
        for predefined_pset in predefined_psets:
            add_property_set(predefined_pset, xml_grouping)

    def add_objects() -> None:

        def add_object(obj: classes.Object, xml_parent) -> None:
            def add_ifc_mapping():
                xml_ifc_mappings = etree.SubElement(xml_object, constants.IFC_MAPPINGS)
                for mapping in obj.ifc_mapping:
                    xml_ifc_mapping = etree.SubElement(xml_ifc_mappings, constants.IFC_MAPPING)
                    xml_ifc_mapping.text = mapping
                pass

            xml_object = etree.SubElement(xml_parent, constants.OBJECT)
            xml_object.set(constants.NAME, obj.name)
            xml_object.set(constants.IDENTIFIER, str(obj.identifier))
            xml_object.set("is_concept", str(obj.is_concept))
            add_parent(xml_object, obj)

            add_ifc_mapping()
            xml_property_sets = etree.SubElement(xml_object, constants.PROPERTY_SETS)
            for property_set in obj.property_sets:
                add_property_set(property_set, xml_property_sets)

            xml_scripts = etree.SubElement(xml_object, constants.SCRIPTS)
            for script in obj.scripts:
                script: classes.Script = script
                xml_script = etree.SubElement(xml_scripts, constants.SCRIPT)
                xml_script.set(constants.NAME, script.name)
                xml_script.text = script.code

        xml_grouping = etree.SubElement(xml_project, constants.OBJECTS)
        for obj in sorted(classes.Object, key=lambda x: x.name):
            add_object(obj, xml_grouping)

    def add_property_set(property_set: classes.PropertySet, xml_parent: etree.Element) -> None:
        def add_attribute(attribute: classes.Attribute, xml_pset: etree._Element) -> None:
            def add_value(attribute: classes.Attribute, xml_attribute: etree._Element) -> None:
                values = attribute.value
                for value in values:
                    xml_value = etree.SubElement(xml_attribute, "Value")
                    if attribute.value_type == constants.RANGE:
                        xml_from = etree.SubElement(xml_value, "From")
                        xml_to = etree.SubElement(xml_value, "To")
                        xml_from.text = str(value[0])
                        if len(value) > 1:
                            xml_to.text = str(value[1])
                    else:
                        xml_value.text = str(value)

            xml_attribute = etree.SubElement(xml_pset, constants.ATTRIBUTE)
            xml_attribute.set(constants.NAME, attribute.name)
            xml_attribute.set(constants.DATA_TYPE, attribute.data_type)
            xml_attribute.set(constants.VALUE_TYPE, attribute.value_type)
            xml_attribute.set(constants.IDENTIFIER, str(attribute.identifier))
            xml_attribute.set(constants.CHILD_INHERITS_VALUE, str(attribute.child_inherits_values))
            xml_attribute.set(constants.REVIT_MAPPING, str(attribute.revit_name))
            add_parent(xml_attribute, attribute)
            obj = attribute.property_set.object
            if obj is not None and attribute == obj.ident_attrib:
                ident = True
            else:
                ident = False

            xml_attribute.set(constants.IS_IDENTIFIER, str(ident))
            add_value(attribute, xml_attribute)

        xml_pset = etree.SubElement(xml_parent, constants.PROPERTY_SET)
        xml_pset.set(constants.NAME, property_set.name)
        xml_pset.set(constants.IDENTIFIER, str(property_set.identifier))
        add_parent(xml_pset, property_set)

        xml_attributes = etree.SubElement(xml_pset, constants.ATTRIBUTES)
        for attribute in property_set.attributes:
            add_attribute(attribute, xml_attributes)

    def add_aggregation(aggregation: classes.Aggregation, xml_nodes: etree.Element) -> None:
        xml_aggregation = etree.SubElement(xml_nodes, constants.NODE)
        xml_aggregation.set(constants.IDENTIFIER, str(aggregation.uuid))
        xml_aggregation.set(constants.OBJECT.lower(), str(aggregation.object.identifier))
        if aggregation.parent is not None:
            xml_aggregation.set(constants.PARENT, str(aggregation.parent.uuid))
        else:
            xml_aggregation.set(constants.PARENT, constants.NONE)
        xml_aggregation.set(constants.IS_ROOT, str(aggregation.is_root))
        if aggregation.parent is not None:
            xml_aggregation.set(constants.CONNECTION, str(aggregation.parent_connection))
        else:
            xml_aggregation.set(constants.CONNECTION, constants.NONE)

    xml_project = etree.Element(constants.PROJECT)
    xml_project.set(constants.NAME, str(project.name))
    xml_project.set(constants.VERSION, str(project.version))
    xml_project.set(constants.AUTHOR, str(project.author))

    add_predefined_property_sets()
    add_objects()

    xml_nodes = etree.SubElement(xml_project, constants.NODES)

    for aggregation in classes.Aggregation:
        add_aggregation(aggregation, xml_nodes)

    tree = etree.ElementTree(xml_project)
    project.reset_changed()
    return tree


def read_xml(project: classes.Project, path: str = False) -> None:
    if not path:
        return
    tree = etree.parse(path)
    projekt_xml = tree.getroot()
    project.author = projekt_xml.attrib.get(constants.AUTHOR)
    project.name = projekt_xml.attrib.get("name")
    project.version = projekt_xml.attrib.get("version")

    def import_property_sets(xml_property_sets: list[etree._Element]) -> (list[classes.PropertySet], classes.Attribute):

        def import_attributes(xml_attributes: etree._Element,
                              property_set: classes.PropertySet) -> classes.Attribute | None:

            def transform_new_values(xml_attribute: etree._Element) -> list[str]:
                def empty_text(xml_value):
                    if xml_value.text is None:
                        return ""
                    else:
                        return xml_value.text

                value_type = xml_attribute.attrib.get("value_type")
                value = list()

                if value_type != constants.RANGE:
                    for xml_value in xml_attribute:
                        value.append(empty_text(xml_value))

                else:
                    for xml_range in xml_attribute:
                        from_to_list = list()
                        for xml_value in xml_range:
                            if xml_value.tag == "From":
                                from_to_list.append(empty_text(xml_value))
                            if xml_value.tag == "To":
                                from_to_list.append(empty_text(xml_value))
                        value.append(from_to_list)
                return value

            ident_attrib = None

            for xml_attribute in xml_attributes:
                attribs = xml_attribute.attrib
                name = attribs.get(constants.NAME)
                identifier = attribs.get(constants.IDENTIFIER)
                data_type = attribs.get(constants.DATA_TYPE)
                value_type = attribs.get(constants.VALUE_TYPE)
                is_identifier = attribs.get(constants.IS_IDENTIFIER)
                child_inh = string_to_bool(attribs.get(constants.CHILD_INHERITS_VALUE))
                value = transform_new_values(xml_attribute)
                attrib = classes.Attribute(property_set, name, value, value_type, data_type, child_inh, identifier)
                revit_mapping = attribs.get(constants.REVIT_MAPPING)
                if revit_mapping == constants.NONE:
                    revit_mapping = attrib.name
                attrib.revit_name = revit_mapping
                if is_identifier == str(True):
                    ident_attrib = attrib
            return ident_attrib

        property_sets: list[classes.PropertySet] = list()
        ident_attrib = None

        for xml_property_set in xml_property_sets:
            attribs = xml_property_set.attrib
            name = attribs.get(constants.NAME)
            identifier = attribs.get(constants.IDENTIFIER)
            property_set = classes.PropertySet(name, obj=None, identifier=identifier)

            xml_attrib_group = xml_property_set.find(constants.ATTRIBUTES)
            ident_value = import_attributes(xml_attrib_group, property_set)
            if ident_value is not None:
                ident_attrib = ident_value
            property_sets.append(property_set)

        return property_sets, ident_attrib

    def import_objects(xml_objects: list[etree._Element]):

        def get_obj_data(xml_object: etree._Element) -> (str, str, str, bool):

            obj_name: str = xml_object.attrib.get(constants.NAME)
            obj_parent: str = xml_object.attrib.get(constants.PARENT)
            identifier: str = xml_object.attrib.get(constants.IDENTIFIER)
            obj_is_concept: str = xml_object.attrib.get(constants.IS_CONCEPT)

            return obj_name, obj_parent, identifier, string_to_bool(obj_is_concept)

        def import_scripts(xml_scripts: etree.Element | None, obj: classes.Object) -> None:
            if xml_scripts is None:
                return
            for xml_script in xml_scripts:
                name = xml_script.attrib.get("name")
                code = xml_script.text
                script = classes.Script(name, obj)
                script.code = code

        for xml_object in xml_objects:
            xml_property_group = xml_object.find(constants.PROPERTY_SETS)
            xml_script_group = xml_object.find(constants.SCRIPTS)
            xml_mapping_group = xml_object.find(constants.IFC_MAPPINGS)

            property_sets, ident_attrib = import_property_sets(xml_property_group)
            name, parent, identifer, is_concept = get_obj_data(xml_object)
            obj = classes.Object(name, ident_attrib, identifier=identifer)
            ident_dict[identifer] = obj

            obj.ifc_mapping = {mapping.text for mapping in xml_mapping_group}
            if not obj.ifc_mapping:
                obj.ifc_mapping.add("IfcBuildingElementProxy")

            for property_set in property_sets:
                obj.add_property_set(property_set)

            import_scripts(xml_script_group, obj)

    def create_ident_dict(item_list: list[Type[classes.Hirarchy]]) -> dict[str, Type[classes.Hirarchy]]:
        return {item.identifier: item for item in item_list}

    def link_parents(xml_predefined_psets: list[etree._Element], xml_objects: list[etree._Element]) -> None:
        def fill_dict(xml_dict: dict[str, str], xml_obj: etree._Element) -> None:
            xml_dict[xml_obj.attrib.get(constants.IDENTIFIER)] = xml_obj.attrib.get(constants.PARENT)

        def iterate() -> None:
            for xml_predefined_pset in xml_predefined_psets:
                fill_dict(xml_property_set_dict, xml_predefined_pset)
                xml_attributes = xml_predefined_pset.find(constants.ATTRIBUTES)
                for xml_attribute in xml_attributes:
                    fill_dict(xml_attribute_dict, xml_attribute)

            for xml_object in xml_objects:
                fill_dict(xml_object_dict, xml_object)
                xml_property_sets = xml_object.find(constants.PROPERTY_SETS)
                for xml_property_set in xml_property_sets:
                    fill_dict(xml_property_set_dict, xml_property_set)
                    xml_attributes = xml_property_set.find(constants.ATTRIBUTES)
                    for xml_attribute in xml_attributes:
                        uuid = xml_attribute.attrib["identifer"]
                        if xml_attribute_dict.get(uuid) is not None:
                            print(f"ERROR DUPLICATED UUID {uuid}")
                        fill_dict(xml_attribute_dict, xml_attribute)

        def create_link(item_dict: dict[str, Type[classes.Hirarchy]], xml_dict: dict[str, str]):
            for ident, item in item_dict.items():
                parent_ident = xml_dict[str(ident)]
                parent_item = item_dict.get(parent_ident)
                if parent_item is not None:
                    parent_item.add_child(child=item)

        xml_property_set_dict = dict()
        xml_object_dict = dict()
        xml_attribute_dict = dict()
        iterate()

        obj_dict = create_ident_dict(classes.Object)
        property_set_dict = create_ident_dict(classes.PropertySet)
        attribute_dict = create_ident_dict(classes.Attribute)

        create_link(obj_dict, xml_object_dict)
        create_link(property_set_dict, xml_property_set_dict)
        create_link(attribute_dict, xml_attribute_dict)

    def link_aggregation() -> None:
        def create_node_dict() -> dict[str, [classes.Aggregation, object]]:

            id_node_dict = dict()
            for xml_node in xml_group_nodes:
                identifier = xml_node.attrib.get(constants.IDENTIFIER)
                obj = ident_dict[xml_node.attrib.get(constants.OBJECT.lower())]
                aggregation = classes.Aggregation(obj, identifier)
                id_node_dict[identifier] = (aggregation, xml_node)
            return id_node_dict

        id_node_dict = create_node_dict()

        for identifier, (aggregation, xml_node) in id_node_dict.items():
            parent_id = xml_node.attrib.get(constants.PARENT)
            is_root = xml_node.attrib.get(constants.IS_ROOT)
            connection_type = xml_node.attrib.get(constants.CONNECTION)
            if parent_id != constants.NONE:
                parent_node: classes.Aggregation = id_node_dict[parent_id][0]
                parent_node.add_child(aggregation, int(connection_type))

    xml_group_predef_psets = projekt_xml.find(constants.PREDEFINED_PSETS)
    xml_group_objects = projekt_xml.find(constants.OBJECTS)
    xml_group_nodes = projekt_xml.find(constants.NODES)

    import_property_sets(xml_group_predef_psets)
    ident_dict: dict[str, classes.Object] = dict()
    import_objects(xml_group_objects)

    link_parents(xml_group_predef_psets, xml_group_objects)
    link_aggregation()


def create_mapping_script(project:classes.Project,pset_name:str,path:str):
    attrib_dict = dict()
    obj: classes.Object
    for obj in project.objects:
        klass = obj.ident_attrib.value[0]
        obj_dict = dict()
        for pset in obj.property_sets:
            pset_dict = dict()
            for attribute in pset.attributes:
                name = attribute.name
                data_format = attribute.data_type
                pset_dict[name] = data_format
            obj_dict[pset.name] = pset_dict
        attrib_dict[klass] = obj_dict
    file_loader = jinja2.FileSystemLoader(HOME_DIR)
    env = jinja2.Environment(loader=file_loader)
    env.trim_blocks = True
    env.lstrip_blocks = True

    template = env.get_template(MAPPING_TEMPLATE)
    code = template.render(attribute_dict=attrib_dict,pset_name = pset_name)
    with open(path, "w") as file:
        file.write(code)
    pass


from __future__ import annotations

import copy
from typing import Iterator
from uuid import uuid4
from anytree import AnyNode
from . import constants, filehandling
from .external_software import excel

# Add child to Parent leads to reverse

class IterRegistry(type):
    _registry = set()
    """ Helper for Iteration"""

    def __iter__(self) -> Iterator[PropertySet | Object | Attribute | Aggregation]:
        return iter(sorted(list(self._registry), key=lambda x: x.name))

    def __len__(self) -> int:
        return len(self._registry)


class Project(object):
    def __init__(self, name: str, author: str = None) -> None:
        self._name = ""
        self._author = author
        self._version = "1.0.0"
        self._changed = True
        self.name = name
        self.seperator_status = True
        self.seperator = ","

    @property
    def changed(self) -> bool:
        def check_data():
            for obj in Object:
                if obj.changed:
                    return True
            return False

        data = check_data()
        if data or self._changed:
            self._changed = True
        else:
            self._changed = False

        return self._changed

    @changed.setter
    def changed(self, value: bool) -> None:
        self._changed = value

    def reset_changed(self) -> None:
        for obj in Object:
            obj.changed = False
        self._changed = False

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        self._changed = True

    @property
    def author(self) -> str:
        return self._author

    @author.setter
    def author(self, value: str):
        self._author = value
        self._changed = True

    @property
    def version(self) -> str:
        return self._version

    @version.setter
    def version(self, value: str):
        self._version = value
        self._changed = True

    def save(self, path: str):
        tree = filehandling.build_xml(self)
        with open(path, "wb") as f:
            tree.write(f, pretty_print=True, encoding="utf-8", xml_declaration=True)

    def clear(self):
        for obj in Object:
            obj.delete()
        for pset in PropertySet:
            pset.delete()

        for attribute in Attribute:
            attribute.delete()
        self.name = ""
        self.author = ""
        self.version = "1.0.0"
        self.changed = True
        self.name = ""
        self.seperator_status = True
        self.seperator = ","

    def open(self, path:str) -> None:
        filehandling.read_xml(self, path)

    def import_excel(self, path:str,ws_name:str) -> None:
        excel.open_file(path,ws_name)

    @property
    def objects(self) -> Iterator[Object]:
        return iter(Object)

    def tree(self) -> AnyNode:
        def create_childen(node: AnyNode):
            obj: Object = node.obj
            for child in obj.children:
                child_node = AnyNode(id=child.name, obj=child, parent=node)
                create_childen(child_node)

        base = AnyNode(id=self.name, obj=self)
        root_objects = [AnyNode(id=obj.name, obj=obj, parent=base) for obj in Object if obj.parent is None]
        for n in root_objects:
            create_childen(n)
        return base

class Hirarchy(object, metaclass=IterRegistry):

    def __init__(self, name: str) -> None:

        self._parent = None
        self._children = set()
        self._name = name
        self.changed = True
        self._mapping_dict = {
            constants.SHARED_PARAMETERS: True,
            constants.IFC_MAPPING: True
        }

    @property
    def mapping_dict(self) -> dict[str, bool]:
        return self._mapping_dict

    @mapping_dict.setter
    def mapping_dict(self, value: dict[str, bool]) -> None:
        self._mapping_dict = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        for child in self.children:
            child.name = value
        self.changed = True

    @property
    def parent(self) -> Hirarchy:
        return self._parent

    @parent.setter
    def parent(self, parent: Hirarchy) -> None:
        self._parent = parent
        self.changed = True

    @property
    def is_parent(self) -> bool:
        if self.children:
            return True
        else:
            return False

    @property
    def is_child(self) -> bool:
        if self.parent is not None:
            return True
        else:
            return False

    @property
    def children(self) -> set[Hirarchy]:
        return self._children

    def add_child(self, child: Hirarchy) -> None:
        self.children.add(child)
        child.parent = self
        self.changed = True

    def remove_child(self, child: Hirarchy) -> None:
        self.children.remove(child)
        child.delete()

    def delete(self) -> None:
        if self in self._registry:
            self._registry.remove(self)


class PropertySet(Hirarchy):
    _registry: set[PropertySet] = set()

    def __init__(self, name: str, obj: Object = None, identifier: str = None) -> None:
        super(PropertySet, self).__init__(name)
        self._attributes = set()
        self._object = obj
        self._registry.add(self)
        self.identifier = identifier
        if self.identifier is None:
            self.identifier = str(uuid4())
        self.changed = True

    @property
    def is_predefined(self) -> bool:
        if self.object is None:
            return True
        else:
            return False

    @property
    def parent(self) -> PropertySet:
        parent = super(PropertySet, self).parent
        return parent

    @parent.setter
    def parent(self, parent: PropertySet) -> None:
        if parent is None:
            self.remove_parent(self._parent)
            return
        self._parent = parent

    def change_parent(self, new_parent: PropertySet) -> None:
        for attribute in self.attributes:
            if attribute.parent.property_set == self._parent:
                self.remove_attribute(attribute)
        self.parent = new_parent

    def delete(self) -> None:
        super(PropertySet, self).delete()
        if self.object is not None:
            ident = self.object.ident_attrib  # if identifier in Pset delete all attributes except identifier
            if ident in self.attributes:
                remove_list = [attribute for attribute in self.attributes if attribute != ident]
                for attribute in remove_list:
                    self.remove_attribute(attribute)
            else:
                self.object.remove_property_set(self)

    @property
    def object(self) -> Object:
        return self._object

    @object.setter
    def object(self, value: Object):
        self._object = value
        self.changed = True

    @property
    def attributes(self) -> set[Attribute]:
        return self._attributes

    @attributes.setter
    def attributes(self, value: set[Attribute]) -> None:
        self._attributes = value
        self.changed = True

    def add_attribute(self, value: Attribute) -> None:
        self._attributes.add(value)
        self.changed = True

        # if value.property_set is not None:
        #     value.property_set.remove_attribute(value)
        value._property_set = self
        for child in self.children:
            attrib: Attribute = copy.copy(value)
            value.add_child(attrib)
            child.add_attribute(attrib)

    def remove_attribute(self, value: Attribute) -> None:
        if value in self.attributes:
            self.attributes.remove(value)
            for child in self.children:
                for attribute in list(child.attributes):
                    if attribute.parent == value:
                        child.remove_attribute(attribute)
            self.changed = True

    def get_attribute_by_name(self, name: str):
        for attribute in self.attributes:
            if attribute.name.lower() == name.lower():
                return attribute
        return None

    def remove_parent(self, old_parent: PropertySet):
        remove_list = list()
        for attribute in self.attributes:
            if attribute.parent.property_set == old_parent:
                remove_list.append(attribute)

        for attribute in remove_list:
            self.remove_attribute(attribute)
        self._parent = None

    def __copy__(self):
        new_pset = PropertySet(self.name)

        for attribute in self.attributes:
            new_attrib = copy.copy(attribute)
            new_pset.add_attribute(new_attrib)

        if self.parent is not None:
            self.parent.add_child(new_pset)

        return new_pset

    def create_child(self, name) -> PropertySet:
        child = PropertySet(name)
        self.children.add(child)
        child.parent = self
        for attribute in self.attributes:
            new_attrib = attribute.create_child()
            child.add_attribute(new_attrib)
        return child


class Attribute(Hirarchy):
    _registry: set[Attribute] = set()

    def __init__(self, property_set: PropertySet | None, name: str, value: list, value_type: int,
                 data_type: str = "xs:string",
                 child_inherits_values: bool = False, identifier: str = None):

        super(Attribute, self).__init__(name=name)
        self._value = value
        self._property_set = property_set
        self._value_type = value_type
        self._data_type = data_type
        self._registry.add(self)
        self._revit_name = name

        self.changed = True
        self._child_inherits_values = child_inherits_values
        self.identifier = identifier
        if self.identifier is None:
            self.identifier = str(uuid4())
        if property_set is not None:
            property_set.add_attribute(self)

    def __str__(self) -> str:
        text = f"{self.property_set.name} : {self.name} = {self.value}"
        return text

    @property
    def revit_name(self) -> str:
        return self._revit_name

    @revit_name.setter
    def revit_name(self, value: str) -> None:
        self._revit_name = value

    @property
    def child_inherits_values(self) -> bool:
        return self._child_inherits_values

    @child_inherits_values.setter
    def child_inherits_values(self, value: bool) -> None:
        self._child_inherits_values = value

    @property
    def name(self) -> str:
        return super(Attribute, self).name

    @name.setter
    def name(self, value: str) -> None:
        self.changed = True  # ToDo: add request for unlink
        self._name = value
        for child in self.children:
            child.name = value

    @property
    def value(self) -> list:
        return self._value

    @value.setter
    def value(self, value: list) -> None:
        def can_be_changed() -> bool:
            change_bool = True
            if self.is_child:
                parent: Attribute = self.parent
                if parent.child_inherits_values:
                    change_bool = False
            return change_bool

        new_value = []

        for el in value:
            if isinstance(el, str):
                if "|" in el:
                    el = el.split("|")
                    for item in el:
                        new_value.append(item)
                else:
                    new_value.append(el)
            else:
                new_value.append(el)

        if can_be_changed():
            self._value = new_value
            self.changed = True

    @property
    def value_type(self) -> int:
        return self._value_type

    @value_type.setter
    def value_type(self, value: int):

        if not self.is_child:
            self._value_type = value
            self.changed = True

        if self.is_parent:
            for child in self.children:
                child._value_type = value
                self.changed = True

    @property
    def data_type(self) -> str:
        """
        "xs:string"; "xs:double"; "xs:boolean"; XS_INT = "xs:int"
        :return:
        """

        return self._data_type

    @data_type.setter
    def data_type(self, value: str) -> None:
        if not self.is_child:
            self._data_type = value
            self.changed = True

        if self.is_parent:
            for child in self.children:
                child._data_type = value
                self.changed = True

    @property
    def property_set(self) -> PropertySet:
        return self._property_set

    def is_equal(self, attribute: Attribute) -> bool:
        equal = True

        if self.name != attribute.name:
            equal = False

        if self.value != attribute.value:
            equal = False

        if self.property_set.name != attribute.property_set.name:
            equal = False

        if equal:
            return True

    def delete(self) -> None:
        self.property_set.remove_attribute(self)
        for child in self.children:
            child.delete()

    def create_child(self) -> Attribute:
        child = copy.copy(self)
        self.add_child(child)
        return child

    def __copy__(self) -> Attribute:
        new_attrib: Attribute = Attribute(None, self.name, self.value,
                                          self.value_type, self.data_type, self.child_inherits_values)
        if self.parent is not None:
            self.parent.add_child(new_attrib)
        return new_attrib


class Object(Hirarchy):
    _registry: set[Object] = set()

    def __init__(self, name: str, ident_attrib: [Attribute, str], identifier: str = None,
                 ifc_mapping: set[str] | None = None) -> None:
        super(Object, self).__init__(name=name)
        self._registry.add(self)

        self._scripts: list[Script] = list()
        self._property_sets: list[PropertySet] = list()
        self._ident_attrib = ident_attrib
        self._aggregations: set[Aggregation] = set()

        if ifc_mapping is None:
            self._ifc_mapping = {"IfcBuildingElementProxy"}
        else:
            self._ifc_mapping = ifc_mapping

        self.changed = True

        if identifier is None:
            self.identifier = str(uuid4())
        else:
            self.identifier = identifier

    def __str__(self):
        return f"Object {self.name}"

    @property
    def ifc_mapping(self) -> set[str]:
        return self._ifc_mapping

    @ifc_mapping.setter
    def ifc_mapping(self, value: set[str]):
        value_set = set()
        for item in value:  # filter empty Inputs
            if not (item == "" or item is None):
                value_set.add(item)
        self._ifc_mapping = value_set

    def add_ifc_map(self, value: str) -> None:
        self._ifc_mapping.add(value)

    def remove_ifc_map(self, value: str) -> None:
        self._ifc_mapping.remove(value)

    @property
    def aggregation_representations(self) -> set[Aggregation]:  # Todo: add nodes functionality to graphs_window
        return self._aggregations

    def add_aggregation_representation(self, node: Aggregation) -> None:
        self._aggregations.add(node)

    def remove_node(self, node: Aggregation) -> None:
        self.aggregation_representations.remove(node)

    @property
    def inherited_property_sets(self) -> dict[Object, list[PropertySet]]:
        def recursion(recursion_property_sets, recursion_obj: Object):
            psets = recursion_obj.property_sets

            if psets:
                recursion_property_sets[recursion_obj] = psets

            parent = recursion_obj.parent
            if parent is not None:
                recursion_property_sets = recursion(recursion_property_sets, parent)
            return recursion_property_sets

        property_sets = dict()
        if self.parent is not None:
            inherited_property_sets = recursion(property_sets, self.parent)
        else:
            inherited_property_sets = dict()

        return inherited_property_sets

    @property
    def is_concept(self) -> bool:
        if isinstance(self.ident_attrib, Attribute):
            return False
        else:
            return True

    @property
    def ident_attrib(self) -> Attribute:
        return self._ident_attrib

    @ident_attrib.setter
    def ident_attrib(self, value: Attribute) -> None:
        self._ident_attrib = value
        self.changed = True

    @property
    def property_sets(self) -> list[PropertySet]:
        return self._property_sets

    # override name setter because of intheritance
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        self.changed = True

    def add_property_set(self, property_set: PropertySet) -> None:
        self._property_sets.append(property_set)
        property_set.object = self

    def remove_property_set(self, property_set: PropertySet) -> None:
        if property_set in self._property_sets:
            self._property_sets.remove(property_set)

    def get_attributes(self, inherit: bool = False) -> list[Attribute]:
        attributes = list()
        for property_set in self.property_sets:
            attributes += property_set.attributes

        if inherit:
            attributes += self.parent.get_attributes(inherit=True)

        return attributes

    @property
    def scripts(self) -> list[Script]:
        return self._scripts

    def add_script(self, script: Script) -> None:
        self._scripts.append(script)

    def delete_script(self, script: Script) -> None:
        self._scripts.remove(script)

    def delete(self) -> None:
        Object._registry.remove(self)
        pset: PropertySet
        for pset in self.property_sets:
            pset.delete()

        for aggregation in self.aggregation_representations.copy():
            aggregation.delete()

    def get_property_set_by_name(self, property_set_name: str) -> PropertySet | None:
        for property_set in self.property_sets:
            if property_set.name == property_set_name:
                return property_set
        return None

    @property
    def ident_value(self):
        if not self.is_concept:
            return self.ident_attrib.value[0]
        return None


class Aggregation(Hirarchy):
    _registry: set[Aggregation] = set()

    def __str__(self):
        return self.name

    def __init__(self, obj: Object, uuid: str | None = None):
        super(Aggregation, self).__init__(name=obj.name)
        self._registry.add(self)
        if uuid is None:
            self.uuid = str(uuid4())
        else:
            self.uuid = str(uuid)
        self.object = obj
        self._parent: Aggregation | None = None
        self.connection_dict: dict[Aggregation, int] = dict()

        self.object.add_aggregation_representation(self)

    @property
    def parent_connection(self) -> int:
        """
        NO PARENT = 0
        AGGREGATION = 1
        INHERITANCE = 2
        AGGREGATION+INHERITANCE =3
        :return:
        """
        if self.parent is None:
            return 0

        return self.parent.connection_dict[self]

    def add_child(self, child: Aggregation, connection_type: int = constants.AGGREGATION) -> Aggregation:
        self.children.add(child)
        child.set_parent(self, connection_type)
        self.connection_dict[child] = connection_type
        return child

    @property
    def parent(self) -> Aggregation:
        return self._parent

    def set_parent(self, value: Aggregation, connection_type: int) -> None:
        if self.parent is not None:
            self.connection_dict.pop(self.parent)
        self._parent = value
        self.connection_dict[value] = connection_type

    @property
    def is_root(self):
        return not self.parent


class Script():
    def __init__(self, title: str, obj: Object) -> None:
        self.code = str()
        self.changed = True
        self._object = obj
        obj.add_script(self)
        self._name = title

    @property
    def object(self) -> Object:
        return self._object

    @object.setter
    def object(self, value: Object) -> None:
        self._object.delete_script(self)
        self._object = value
        value.add_script(self)
        self.changed = True

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        self.changed = True



import os

import openpyxl
from lxml import etree, builder
from openpyxl import load_workbook,Workbook
from openpyxl.worksheet.worksheet import  Worksheet

from .. import classes


def _create_sheet(obj:classes.Object, workbook:Workbook,name):
    sheet = workbook.create_sheet(name)
    attributes = set()
    for property_set in obj.property_sets:
        for attribute in property_set.attributes:
            attributes.add(attribute.name)
    for i,attrib_name in enumerate(sorted(attributes),start=1):
        sheet.cell(1,i).value = attrib_name


def create_mapping(src_path: str, dest_path: str, project:classes.Project) -> None:


    export_wb = Workbook()
    export_wb.active.title="Hilfe"
    wb = load_workbook(src_path)
    sheet:Workbook = wb.active
    important_rows = [row for i, row in enumerate(sheet.rows) if row[2].value is not None and i != 0]
    object_dict = {obj.ident_attrib.value[0]: obj for obj in project.objects if not obj.is_concept}

    for row in important_rows:
        bauteil_bez_card, bauteil_bez_2, bauteilklass = map(lambda x: x.value, row)
        obj = object_dict.get(bauteilklass)
        if obj is None:
            print(f"BauteilKlassifikation '{bauteilklass}' not found")
            continue
        _create_sheet(obj,export_wb,bauteil_bez_card)
    export_wb.save(dest_path)



import os
import importlib.resources
import json

HOME_DIR = os.path.dirname(__file__)
TEMPLATE = "template.txt"
MAPPING_TEMPLATE = "mapping_template.txt"


with importlib.resources.open_text("SOMcreator.Template","ifc.json") as file:
    dict = json.load(file)
    IFC_4_1 = [entity.strip() for entity in dict["ifc4.1"]]


from __future__ import annotations

import logging
import os.path
import shutil
import tempfile
import re
import json
from typing import Iterator

import openpyxl
from openpyxl.cell.cell import Cell
from openpyxl.worksheet.worksheet import Worksheet

from .. import classes, constants
from ..Template import IFC_4_1

ident_pset_name = "Allgemeine Eigenschaften"
ident_attrib_name = "bauteilKlassifikation"


def split_string(text: str) -> list[str] | None:
    if text is None:
        return []
    text = re.findall(r'[^,;]+', text)
    for i, item in enumerate(text):
        if "(" in item:
            item = item.split("(")
            text[i] = item[0]
        text[i] = text[i].strip()

    return text


def find_by_abbreviation(abbreviation: str) -> ExcelBlock | None:
    excel_block_dict: dict[str, ExcelBlock] = {block.abbreviation.upper(): block for block in ExcelBlock}
    return excel_block_dict.get(abbreviation.upper())


class ExcelIterator(type):
    def __iter__(cls: ExcelBlock) -> Iterator[ExcelBlock]:
        return iter(cls._registry)


class ExcelBlock(metaclass=ExcelIterator):
    _registry: list[ExcelBlock] = list()

    def __init__(self, base_cell: Cell, sheet: Worksheet):
        self._registry.append(self)
        self.base_cell = base_cell
        self.sheet = sheet
        self.pset: classes.PropertySet | None = None
        self.object: classes.Object | None = None

    @property
    def ident_value(self) -> str | None:
        cell = self.sheet.cell(row=self.base_cell.row,
                               column=self.base_cell.column + 2)
        return cell.value

    @property
    def name(self) -> str:
        cell = self.sheet.cell(row=self.base_cell.row,
                               column=self.base_cell.column + 1)
        return cell.value

    @property
    def abbreviation(self) -> str:
        cell = self.sheet.cell(row=self.base_cell.row + 1,
                               column=self.base_cell.column + 1)
        return cell.value.upper()

    @property
    def is_predefined_pset(self) -> bool:
        if self.ident_value is None:
            return True
        else:
            return False

    @property
    def entry(self) -> Cell:
        if self.is_predefined_pset:
            return self.sheet.cell(row=self.base_cell.row + 4, column=self.base_cell.column)
        else:
            return self.sheet.cell(row=self.base_cell.row + 5, column=self.base_cell.column)

    @property
    def aggregates(self) -> set[ExcelBlock] | None:
        if self.is_predefined_pset:
            return None

        child_cell = self.sheet.cell(row=self.base_cell.row + 3, column=self.base_cell.column + 1)
        child_string_list = split_string(child_cell.value)
        excel_block_dict: dict[str, ExcelBlock] = {block.abbreviation: block for block in ExcelBlock}

        children = set()
        for abbrev in child_string_list:
            if abbrev != "-":
                block = find_by_abbreviation(abbrev)
                if block is not None:
                    children.add(block)
                else:
                    logging.error(f"[{self.name}] abbreviation '{abbrev}' doesn't exist!")
        return children

    @property
    def parent_classes(self) -> set[ExcelBlock] | None:
        parents = set()

        parent_text = self.sheet.cell(self.base_cell.row + 2, self.base_cell.column + 1).value
        for abbrev in split_string(parent_text):
            if abbrev != "-":
                parent_block = find_by_abbreviation(abbrev)
                if parent_block is not None:
                    if parent_block.name != ident_pset_name:
                        parents.add(parent_block)
                        parents = set.union(parent_block.parent_classes, parents)
                else:

                    logging.warning(
                        f"[{self.name}] Elternklasse: Kürzel {abbrev.upper()} existiert nicht!")
        return parents

    @property
    def inherits(self) -> set[ExcelBlock]:

        def is_child(value) -> bool:
            last_num = value.split(".")[-1]

            if not last_num.isdigit():
                return False
            last_num = int(last_num)
            if last_num >= 100 and last_num % 100 == 0:
                return True
            else:
                return False

        inherit_blocks = set()

        if self.is_predefined_pset:
            return inherit_blocks

        for block in ExcelBlock:
            if not block.is_predefined_pset:
                parent_txt = block.ident_value.split(".")[:-1]
                parent_txt = ".".join(parent_txt)
                if parent_txt == self.ident_value and is_child(block.ident_value):
                    inherit_blocks.add(block)

        return inherit_blocks

    def create_attributes(self) -> list[classes.Attribute]:
        """
           Iterate over Attributes
           Create Them and find special Datatypes
           """

        def transform_value_types(value: str) -> (str, bool):
            special = False
            if value is not None:
                if value.lower() in ["string", "str"]:
                    data_type = constants.XS_STRING
                elif value.lower() in ["double"]:
                    data_type = constants.XS_DOUBLE
                elif value.lower() in ["boolean", "bool"]:
                    data_type = constants.XS_BOOL
                elif value.lower() in ["int", "integer"]:
                    data_type = constants.XS_INT
                else:
                    special = True
                    data_type = constants.XS_STRING
            else:
                data_type = constants.XS_STRING

            return data_type, special

        if self.is_predefined_pset:
            row = self.base_cell.row + 4
        else:
            row = self.base_cell.row + 5

        entry = self.sheet.cell(row=row, column=self.base_cell.column)

        cell_list = [block.base_cell for block in ExcelBlock]
        attributes: list[classes.Attribute] = list()

        while entry.value is not None and entry not in cell_list and entry.value != "-":
            data_type_text = self.sheet.cell(row=entry.row, column=entry.column + 2).value
            data_type, special = transform_value_types(data_type_text)
            attribute_name = entry.value
            alternative_name = self.sheet.cell(row=entry.row, column=entry.column + 1).value
            attribute = classes.Attribute(self.pset, attribute_name, [], constants.VALUE_TYPE_LOOKUP[constants.LIST],
                                          data_type=data_type)
            if alternative_name and alternative_name is not None:
                attribute.revit_name = alternative_name
            attributes.append(attribute)

            if special:
                logging.info(
                    f"[{entry.value}] Property: {self.name}:{attribute_name} datatype '{data_type_text}' unbekannt")
                pass

            entry = self.sheet.cell(row=entry.row + 1, column=entry.column)

        return attributes

    def create_predefined_pset(self) -> None:
        self.pset = classes.PropertySet(self.name)
        self.pset.attributes = self.create_attributes()

    def create_object(self) -> classes.Object:
        self.pset = classes.PropertySet(self.name)
        self.pset.attributes = self.create_attributes()
        predef_psets: dict[str, ExcelBlock] = {block.name: block for block in ExcelBlock if block.is_predefined_pset}
        parent_pset = predef_psets.get(ident_pset_name)

        ident_pset = parent_pset.pset.create_child(ident_pset_name)
        ident_attrib = ident_pset.get_attribute_by_name(ident_attrib_name)
        ident_attrib.value = [self.ident_value]
        obj = classes.Object(self.name, ident_attrib)
        obj.add_property_set(self.pset)
        obj.add_property_set(ident_pset)
        obj.ifc_mapping = self.ifc_mapping()
        return obj

    def ifc_mapping(self) -> set[str]:
        if self.is_predefined_pset:
            return set()

        cell = self.sheet.cell(row=self.base_cell.row + 4, column=self.base_cell.column + 2)
        value: str = cell.value
        if value is None:
            logging.error(f"[{self.name}]: no IFC Mapping")
            return {"IfcBuildingElementProxy"}

        string_list = value.split("/")
        string_list = [string.strip() for string in string_list]

        for string in string_list:
            if string not in IFC_4_1:
                logging.info(f"[{self.name}]: '{string}' not in IFC 4.1 Specification")
        return set(string_list)

def _create_blocks(sheet) -> None:
    """create Excel Blocks"""
    excel_blocks = set()

    row: tuple[Cell]
    for row in sheet:
        for cell in row:
            if cell.value is not None:
                text = cell.value.strip()
                if text in ["name", "name:"]:
                    if sheet.cell(cell.row + 1, cell.column).value == "Kürzel":
                        excel_blocks.add(ExcelBlock(cell, sheet))
                    else:
                        logging.error(f"{sheet.cell(cell.row + 1, cell.column)} hat den Wert 'name'")

def _create_items() -> None:
    """create Objects and PropertySets"""
    predef_psets = [block for block in ExcelBlock if block.is_predefined_pset]
    objects = [block for block in ExcelBlock if not block.is_predefined_pset]

    for predef_pset_block in predef_psets:
        predef_pset_block.create_predefined_pset()

    for object_block in objects:
        obj = object_block.create_object()
        object_block.object = obj

    for block in objects:
        for pset in [block.pset for block in block.parent_classes]:
            new_pset = pset.create_child(pset.name)
            block.object.add_property_set(new_pset)

def _build_object_tree() -> None:
    tree_dict: dict[str, classes.Object] = {obj.ident_attrib.value[0]: obj for obj in classes.Object}

    for ident, item in tree_dict.items():
        ident_list = ident.split(".")[:-1]
        parent_obj = tree_dict.get(".".join(ident_list))

        if parent_obj is not None:
            parent_obj.add_child(item)

def _build_aggregations() -> None:
    def get_root_blocks() -> set[ExcelBlock]:
        children = set()
        for e_block in ExcelBlock:  # find all child blocks
            if not e_block.is_predefined_pset:
                children = set.union(e_block.aggregates, children)
                children = set.union(e_block.inherits, children)

        r_blocks = set()  # root blocks
        for e_block in ExcelBlock:
            if e_block not in children and not e_block.is_predefined_pset:
                r_blocks.add(e_block)

        return r_blocks

    def link_child_nodes(aggregation: classes.Aggregation, block: ExcelBlock) -> None:
        aggregate_list = block.aggregates
        inherit_list = block.inherits

        for aggregate_block in aggregate_list:
            if aggregate_block.name != block.name:
                if aggregate_block.is_predefined_pset:
                    logging.error(f"[{block.name}] can't aggregate to {aggregate_block.name}")
                else:
                    child_aggreg = classes.Aggregation(aggregate_block.object)
                    relationship = constants.AGGREGATION
                    if aggregate_block in inherit_list:
                        relationship += constants.INHERITANCE

                    aggregation.add_child(child_aggreg, relationship)
                    link_child_nodes(child_aggreg, aggregate_block)
            else:
                logging.warning(f"[{aggregation.name}] recursive aggregation")

        for inherit_block in inherit_list:
            if inherit_block not in aggregate_list:
                child_aggreg = classes.Aggregation(inherit_block.object)
                aggregation.add_child(child_aggreg, constants.INHERITANCE)
                link_child_nodes(child_aggreg, inherit_block)

    root_blocks = get_root_blocks()

    for block in root_blocks:
        aggregation = classes.Aggregation(block.object)
        link_child_nodes(aggregation, block)

def open_file(path: str,ws_name:str) -> None:

    # TODO: add request for Identification Attribute
    with tempfile.TemporaryDirectory() as tmpdirname:

        new_path = os.path.join(tmpdirname, "excel.xlsx")
        shutil.copy2(path, new_path)
        book = openpyxl.load_workbook(new_path)
        if ws_name not in book.sheetnames:
            logging.error("Worksheet Name not in Workbook")
            return
        sheet = book[ws_name]

        _create_blocks(sheet)
        _create_items()
        _build_object_tree()
        _build_aggregations()

    ExcelBlock._registry = list()

def create_abbreviation_json(excel_path:str,export_path:str,ws_name:str) -> None:
    ExcelBlock._registry = list()
    with tempfile.TemporaryDirectory() as tmpdirname:
        new_path = os.path.join(tmpdirname, "excel.xlsx")
        shutil.copy2(excel_path, new_path)
        book = openpyxl.load_workbook(new_path)
        if ws_name not in book.sheetnames:
            logging.error("Worksheet Name not in Workbook")
            return
        sheet = book[ws_name]

        _create_blocks(sheet)
        _create_items()
        _build_object_tree()
        _build_aggregations()
        print("{")

        d = {block.abbreviation:[block.ident_value,block.name] for block in ExcelBlock if block.ident_value is not None}
        with open(export_path,"w") as file:
            json.dump(d,file,indent=2)